# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2022-04-10

r"""Rekordbox XML database file handler."""

import logging
import os.path
import urllib.parse
import xml.etree.cElementTree as xml
from abc import abstractmethod
from collections import abc

import bidict

from .utils import pretty_xml

logger = logging.getLogger(__name__)

URL_PREFIX = "file://localhost/"
POSMARK_TYPE_MAPPING = bidict.bidict(
    {
        "0": "cue",
        "1": "fadein",
        "2": "fadeout",
        "3": "load",
        "4": "loop",
    }
)
RATING_MAPPING = bidict.bidict({"0": 0, "51": 1, "102": 2, "153": 3, "204": 4, "255": 5})
NODE_KEYTYPE_MAPPING = bidict.bidict({"0": "TrackID", "1": "Location"})


class XmlDuplicateError(Exception):
    """Raised when a track already exists in the XML database."""

    def __init__(self, key_type, key):
        super().__init__(f"XML database already contains a track with {key_type}={key}")


class XmlAttributeKeyError(Exception):
    def __init__(self, cls, key, attributes):
        super().__init__(
            f"{key} is not a valid key for {cls.__name__}! Valid attribs:\n{attributes}"
        )


def encode_path(path):
    r"""Encodes a file path as URI string.

    Parameters
    ----------
    path : str or Path
        The file path to encode.

    Returns
    -------
    url : str
        The encoded file path as URI string.

    Examples
    --------
    >>> s = r"C:\Music\PioneerDJ\Demo Tracks\Demo Track 1.mp3"  # noqa: W605
    >>> encode_path(s)
    file://localhost/C:/Music/PioneerDJ/Demo%20Tracks/Demo%20Track%201.mp3

    """
    url_path = urllib.parse.quote(str(path), safe=":/\\")
    url = URL_PREFIX + url_path.replace("\\", "/")
    return url


def decode_path(url):
    r"""Decodes an as URI string encoded file path.

    Parameters
    ----------
    url : str
        The encoded file path to decode.

    Returns
    -------
    path : str
        The decoded file path.

    Examples
    --------
    >>> s = r"file://localhost/C:/Music/PioneerDJ/Demo%20Tracks/Demo%20Track%201.mp3"
    >>> decode_path(s)
    C:\Music\PioneerDJ\Demo Tracks\Demo Track 1.mp3  # noqa: W605

    """
    path = urllib.parse.unquote(url)
    path = path.replace(URL_PREFIX, "")
    return os.path.normpath(path)


class AbstractElement(abc.Mapping):
    """Abstract base class for Rekordbox XML elements.

    Implements attribute getters and setters for an XML element
    """

    TAG: str
    """str: Name of the XML element"""

    ATTRIBS: list
    """list[str]: List of all attribute keys of the XML element"""

    GETTERS = dict()
    """dict[str, Callable]: Dictionary of attribute getter conversion methods.

    See Also
    --------
    AbstractElement.get
    """
    SETTERS = dict()
    """dict[str, Callable]: Dictionary of attribute setter conversion methods.

    See Also
    --------
    AbstractElement.set
    """

    def __init__(self, element=None, *args, **kwargs):
        self._element = element
        if element is None:
            self._init(*args, **kwargs)
        else:
            self._load_subelements()

    @abstractmethod
    def _init(self, *args, **kwargs):
        """Initializes a new XML element."""
        pass

    def _load_subelements(self):
        """Loads the sub-elements of an existing XML element."""
        pass

    def get(self, key, default=None):
        """Returns the value of an attribute of the XML element.

        The type of the attribute value is converted if a conversion method is specified
        in the ``GETTERS`` class attribute. If no conversion method is found the value
        is returned unconverted as the default type ``str``.

        Parameters
        ----------
        key : str
            The key of the attribute.
        default : Any, optional
            The default value returned if the attribute does not exist.

        Returns
        -------
        value : Any
            The value of the atttribute. The type of the attribute is converted
            acccording to the data of the field.

        Raises
        ------
        XmlAttributeKeyError:
            Raised if `key` is not a valid attribute key.
        """
        if key not in self.ATTRIBS:
            raise XmlAttributeKeyError(self.__class__, key, self.ATTRIBS)
        value = self._element.attrib.get(key, default)
        if value == default:
            return default
        try:
            # Apply callback
            value = self.GETTERS[key](value)
        except KeyError:
            pass
        return value

    def set(self, key, value):
        """Sets the value of an attribute of the XML element.

        The type of the given value is converted before updating the attribute if a
        conversion method is specified in the ``SETTERS`` class attribute.
        If no conversion method is found the value updated set as given.

        Parameters
        ----------
        key : str
            The key of the attribute.
        value : Any
            The value for updating the attribute. The type conversion is handled
            automatically.

        Raises
        ------
        XmlAttributeKeyError:
            Raised if `key` is not a valid attribute key.
        """
        if key not in self.ATTRIBS:
            raise XmlAttributeKeyError(self.__class__, key, self.ATTRIBS)
        try:
            # Apply callback
            value = self.SETTERS[key](value)
        except KeyError:
            # Convert to str just in case
            value = str(value)
        self._element.attrib[key] = value

    def __len__(self):
        """int: The number of attributes of the XML element."""
        return len(self._element.attrib)

    def __iter__(self):
        """Iterable: An iterator of the attribute keys of the XML element."""
        return iter(self._element.attrib.keys())

    def __getitem__(self, key):
        """Returns the raw value of an attribute of the XML element.

        Parameters
        ----------
        key : str
            The key of the attribute.

        Returns
        -------
        value : Any
            The raw value of the attribute.
        """
        return self.get(key)

    def __setitem__(self, key, value):
        """Sets the raw value of an attribute of the XML element.

        Parameters
        ----------
        key : str
            The key of the attribute.
        value : Any
            The raw value for updating the attribute.
        """
        self.set(key, value)

    def __getattr__(self, key):
        """Returns the raw value of an attribute of the XML element (same as `get`)."""
        return self.get(key)

    def __repr__(self):
        return f"<{self.__class__.__name__}()>"


# -- Collection elements ---------------------------------------------------------------


# noinspection PyPep8Naming,PyUnresolvedReferences
class Tempo(AbstractElement):
    """Tempo element representing the beat grid of a track.

    Attributes
    ----------
    Inizio : float
        The start position of the beat grid item.
    Bpm : float
        The BPM value of the beat grid item.
    Metro : str
        The kind of musical meter, for example '4/4'. The default is '4/4'.
    Battito : int
        The beat number in the bar. If `metro` is '4/4', the value can be 1, 2, 3 or 4.
    """

    TAG = "TEMPO"
    ATTRIBS = ["Inizio", "Bpm", "Metro", "Battito"]
    GETTERS = {"Inizio": float, "Bpm": float, "Battito": int}

    def __init__(self, parent=None, Inizio=0.0, Bpm=0.0, Metro="4/4", Battito=1, element=None):
        super().__init__(element, parent, Inizio, Bpm, Metro, Battito)

    def _init(self, parent, inizio, bpm, metro, battito):
        attrib = {
            "Inizio": str(inizio),
            "Bpm": str(bpm),
            "Metro": str(metro),
            "Battito": str(battito),
        }
        self._element = xml.SubElement(parent, self.TAG, attrib=attrib)

    def __repr__(self):
        args = ", ".join(
            [
                f"Inizio={self.Inizio}",
                f"Bpm={self.Bpm}",
                f"Metro={self.Metro}",
                f"Battito={self.Battito}",
            ]
        )
        return f"<{self.__class__.__name__}({args})>"


# noinspection PyPep8Naming,PyUnresolvedReferences
class PositionMark(AbstractElement):
    """Position element for storing position markers like cue points of a track.

    Attributes
    ----------
    Name : str
        The name of the position mark.
    Type : str
        The type of position mark. Can be 'cue', 'fadein', 'fadeout', 'load' or 'loop'.
    Start : float
        Start position of the position mark in seconds.
    End : float, optionl
        End position of the position mark in seconds.
    Num : int, optional
        Charakter for identification of the position mark (for hot cues). For memory
        cues this is always -1.
    """

    TAG = "POSITION_MARK"
    ATTRIBS = ["Name", "Type", "Start", "End", "Num"]

    GETTERS = {
        "Type": POSMARK_TYPE_MAPPING.get,
        "Start": float,
        "End": float,
        "Num": int,
    }
    SETTERS = {"Type": POSMARK_TYPE_MAPPING.inv.get}  # noqa

    def __init__(
        self,
        parent=None,
        Name="",
        Type="cue",
        Start=0.0,
        End=None,
        Num=-1,
        element=None,
    ):
        super().__init__(element, parent, Name, Type, Start, End, Num)

    def _init(self, parent, name, type_, start, end, num):
        attrib = {
            "Name": name,
            "Type": POSMARK_TYPE_MAPPING.inv.get(type_),  # noqa
            "Start": str(start),
            "Num": str(num),
        }
        if end is not None:
            attrib["End"] = str(end)
        self._element = xml.SubElement(parent, self.TAG, attrib=attrib)

    def __repr__(self):
        args = ", ".join(
            [
                f"Name={self.Name}",
                f"Type={self.Type}",
                f"Start={self.Start}",
                f"End={self.End}",
                f"Num={self.Num}",
            ]
        )
        return f"<{self.__class__.__name__}({args})>"


# noinspection PyPep8Naming,PyUnresolvedReferences
class Track(AbstractElement):
    """Track element for storing the metadata of a track.

    Attributes
    ----------
    TrackID : int
        Identification of the track.
    Name: str
        The name of the track.
    Artist : str
        The name of the artist.
    Composer : str
        The name of the composer (or producer).
    Album : str
        The name of the album.
    Grouping : str
        The name of the grouping.
    Genre : str
        The name of the genre.
    Kind : str
        The kind of the audio file, for example 'WAV File' or 'MP3 File'.
    Size : int
        The size of the audio file.
    TotalTime : int
        The duration of the track in seconds.
    DiscNumber : int
        The number of the disc of the album.
    TrackNumber : int
        The Number of the track of the album.
    Year : int
        The year of release.
    AverageBpm : float
        The average BPM of the track.
    DateModified : str
        The date of last modification in the format 'yyyy-mm-dd'.
    DateAdded : str
        The date of addition modification in the format 'yyyy-mm-dd'.
    BitRate : int
        The encoding bit rate.
    SampleRate : float
        The frequency of sampling.
    Comments : str
        The comments of the track.
    PlayCount : int
        The play count of the track.
    LastPlayed : str
        The date of last playing in the format 'yyyy-mm-dd'.
    Rating : int
        The rating of the track using the mapping 0=0, 1=51, 2=102, 3=153, 4=204, 5=255.
    Location : str
        The location of the file encoded as URI string. This value is essential for
        each track.
    Remixer : str
        The name of the remixer.
    Tonality : str
        The tonality or kind of musical key.
    Label : str
        The name of the record label.
    Mix : str
        The name of the mix.
    Colour : str
        The color for track grouping in RGB format.
    tempos : list
        The `Tempo` elements of the track.
    marks : list
        The `PositionMark` elements of the track.

    Raises
    ------
    XmlAttributeKeyError:
        Raised if initialized with invalid key in attributes.
    """

    TAG = "TRACK"
    ATTRIBS = [
        "TrackID",
        "Name",
        "Artist",
        "Composer",
        "Album",
        "Grouping",
        "Genre",
        "Kind",
        "Size",
        "TotalTime",
        "DiscNumber",
        "TrackNumber",
        "Year",
        "AverageBpm",
        "DateModified",
        "DateAdded",
        "BitRate",
        "SampleRate",
        "Comments",
        "PlayCount",
        "LastPlayed",
        "Rating",
        "Location",
        "Remixer",
        "Tonality",
        "Label",
        "Mix",
        "Colour",
    ]

    GETTERS = {
        "TrackID": int,
        "Size": int,
        "TotalTime": int,
        "DiscNumber": int,
        "TrackNumber": int,
        "Year": int,
        "AverageBpm": float,
        "BitRate": int,
        "SampleRate": float,
        "PlayCount": int,
        "Rating": RATING_MAPPING.get,
        "Location": decode_path,
    }

    SETTERS = {"Rating": RATING_MAPPING.inv.get, "Location": encode_path}  # noqa

    def __init__(self, parent=None, Location="", element=None, **kwargs):
        self.tempos = list()
        self.marks = list()
        super().__init__(element, parent, Location, **kwargs)

    def _init(self, parent, Location, **kwargs):
        attrib = {"Location": encode_path(Location)}
        for key, val in kwargs.items():
            if key not in self.ATTRIBS:
                raise XmlAttributeKeyError(self.__class__, key, self.ATTRIBS)
            attrib[key] = str(val)
        self._element = xml.SubElement(parent, self.TAG, attrib=attrib)

    def _load_subelements(self):
        tempo_elements = self._element.findall(f"{Tempo.TAG}")
        if tempo_elements is not None:
            self.tempos = [Tempo(element=el) for el in tempo_elements]
        mark_elements = self._element.findall(f".//{PositionMark.TAG}")
        if mark_elements is not None:
            self.marks = [PositionMark(element=el) for el in mark_elements]

    def add_tempo(self, Inizio, Bpm, Metro, Battito):
        """Adds a new ``Tempo`` XML element to the track element.

        Parameters
        ----------
        Inizio : float
            The start position of the beat grid item.
        Bpm : float
            The BPM value of the beat grid item.
        Metro : str, optional
            The kind of musical meter, for example '4/4'. The default is '4/4'.
        Battito : int
            The beat number in the bar. If `metro` is '4/4', the value can be 1, 2, 3
            or 4.

        Returns
        -------
        tempo : Tempo
            The newly created tempo XML element.

        See Also
        --------
        Tempo: Beat grid XML element handler
        """
        tempo = Tempo(self._element, Inizio, Bpm, Metro, Battito)
        self.tempos.append(tempo)
        return tempo

    def add_mark(self, Name="", Type="cue", Start=0.0, End=None, Num=-1):
        """Adds a new ``PositionMark`` XML element to the track element.

        Parameters
        ----------
        Name : str
            The name of the position mark.
        Type : str
            The type of position mark. Can be 'cue', 'fadein', 'fadeout', 'load' or
            'loop'.
        Start : float
            Start position of the position mark in seconds.
        End : float or None, optionl
            End position of the position mark in seconds.
        Num : int, optional
            Charakter for identification of the position mark (for hot cues). For memory
            cues this is always -1.

        Returns
        -------
        position_mark : PositionMark
            The newly created position mark XML element.

        See Also
        --------
        PositionMark: Position mark XML element handler
        """
        mark = PositionMark(self._element, Name, Type, Start, End, Num)
        self.marks.append(mark)
        return mark

    def __repr__(self):
        return f"<{self.__class__.__name__}(Location={self.Location})>"


# -- Playlist elements -----------------------------------------------------------------


class Node:
    """Node element used for representing playlist folders and playlists.

    A node configured as playlist folder can store other nodes as well as tracks, a node
    configured as playlist can only store tracks. The tracks in playlists are stored via
    a key depending on the key type of the playlist. The key type can either be the
    ID of the track in the XML database ('TrackID') or the file path of the track
    ('Location').
    """

    TAG = "NODE"
    """str: Name of the XML element"""

    FOLDER = 0
    PLAYLIST = 1

    def __init__(self, parent=None, element=None, **attribs):
        self._parent = parent
        self._element = element
        if element is None:
            self._element = xml.SubElement(parent, self.TAG, attrib=attribs)

    @classmethod
    def folder(cls, parent, name):
        """Initializes a playlist folder node XML element.

        Parameters
        ----------
        parent : xml.Element
            The parent node XML element of the new playlist folder node.
        name : str
            The name of the playlist folder node.
        """
        attrib = {"Name": name, "Type": str(cls.FOLDER), "Count": "0"}
        return cls(parent, **attrib)

    @classmethod
    def playlist(cls, parent, name, keytype="TrackID"):
        """Initializes a playlist node XML element.

        Parameters
        ----------
        parent : xml.Element
            The parent node XML element of the new playlist node.
        name : str
            The name of the playlist node.
        keytype : str, optional
            The key type used by the playlist node. Can be 'TrackID' or 'Location'
            (file path of the track).
        """
        attrib = {
            "Name": name,
            "Type": str(cls.PLAYLIST),
            "KeyType": NODE_KEYTYPE_MAPPING.inv[keytype],  # noqa
            "Entries": "0",
        }
        return cls(parent, **attrib)

    @property
    def parent(self):
        """Node: The parent of the node."""
        return self._parent

    @property
    def name(self):
        """str: The name of node."""
        return self._element.attrib.get("Name")

    @property
    def type(self):
        """int: The type of the node (0=folder or 1=playlist)."""
        return int(self._element.attrib.get("Type"))

    @property
    def count(self):
        """int: The number of attributes of the XML element."""
        return int(self._element.attrib.get("Count", 0))

    @property
    def entries(self):
        """int: The number of entries of the node."""
        return int(self._element.attrib.get("Entries", 0))

    @property
    def key_type(self):
        """str: The type of key used by the playlist node."""
        return NODE_KEYTYPE_MAPPING.get(self._element.attrib.get("KeyType"))

    @property
    def is_folder(self):
        """bool: True if the node is a playlist folder, false if otherwise."""
        return self.type == self.FOLDER

    @property
    def is_playlist(self):
        """bool: True if the node is a playlist, false if otherwise."""
        return self.type == self.PLAYLIST

    def _update_count(self):
        self._element.attrib["Count"] = str(len(self._element))

    def _update_entries(self):
        self._element.attrib["Entries"] = str(len(self._element))

    def get_node(self, i):
        """Returns the i-th sub-Node of the current node.

        Parameters
        ----------
        i : int
            Index of sub-Node

        Returns
        -------
        subnode : Node
        """
        return Node(self, element=self._element.findall(f"{self.TAG}[{i + 1}]"))

    def get_playlist(self, name):
        """Returns the sub-Node with the given name.

        Parameters
        ----------
        name : str
            Name of the sub-Node

        Returns
        -------
        subnode : Node
        """
        return Node(self, element=self._element.find(f'.//{self.TAG}[@Name="{name}"]'))

    def get_playlists(self):
        """Returns all sub-nodes that are playlists.

        Returns
        -------
        playlists : list[Node]
            The playlist nodes in the current node.
        """
        return [Node(self, element=el) for el in self._element]

    def add_playlist_folder(self, name):
        """Add a new playlist folder as child to this node.

        Parameters
        ----------
        name : str
            The name of the new playlist folder.

        Returns
        -------
        folder_node : Node
            The newly created playlist folder node.

        Raises
        ------
        ValueError:
            Raised if called on a playlist node.
        """
        if self.is_playlist:
            raise ValueError("Sub-elements can only be added to a folder node!")

        node = Node.folder(self._element, name)
        self._update_count()
        return node

    def add_playlist(self, name, keytype="TrackID"):
        """Add a new playlist as child to this node.

        Parameters
        ----------
        name : str
            The name of the new playlist.
        keytype : {'TrackID', 'Location'} str
            The type of key the playlist uses to store the tracks. Can either be
            'TrackID' or 'Location'.

        Returns
        -------
        playlist_node : Node
            The newly created playlist node.

        Raises
        ------
        ValueError:
            Raised if called on a playlist node.
        """
        if self.is_playlist:
            raise ValueError("Sub-elements can only be added to a folder node!")

        node = Node.playlist(self._element, name, keytype)
        self._update_count()
        return node

    def remove_playlist(self, name):
        """Removes a playlist from the playlist folder node.

        Parameters
        ----------
        name : str
            The name of the playlist to remove.
        """
        item = self.get_playlist(name)
        self._element.remove(item._element)  # noqa
        self._update_count()
        self._update_entries()

    def add_track(self, key):
        """Adds a new track to the playlist node.

        Parameters
        ----------
        key : int or str
            The key of the track to add, depending on the `type` of the playlist node.

        Returns
        -------
        el : xml.SubElement
            The newly created playlist track element.
        """
        el = xml.SubElement(self._element, Track.TAG, attrib={"Key": str(key)})
        self._update_entries()
        return el

    def remove_track(self, key):
        """Removes a track from the playlist node.

        Parameters
        ----------
        key : int or str
            The key of the track to remove, depending on the `type` attribute of the
            playlist node.
        """
        el = self._element.find(f'{Track.TAG}[@Key="{key}"]')
        self._element.remove(el)
        self._update_entries()
        return el

    def get_tracks(self):
        """Returns the keys of all tracks contained in the playlist node.

        Returns
        -------
        keys : list
            The keys of the tracks in the playlist. The format depends on the `type`
            attribute of the playlist node.
        """
        if self.type == self.FOLDER:
            return list()
        elements = self._element.findall(f".//{Track.TAG}")
        items = list()
        for el in elements:
            val = el.attrib["Key"]
            if self.key_type == "TrackID":
                val = int(val)
            items.append(val)
        return items

    def get_track(self, key):
        """Returns the formatted key of the track."""
        el = self._element.find(f'{Track.TAG}[@Key="{key}"]')
        val = el.attrib["Key"]
        if self.key_type == "TrackID":
            val = int(val)
        return val

    def treestr(self, indent=4, lvl=0):
        """returns a formatted string of the node tree strucutre.

        Parameters
        ----------
        indent : int, optional
            Number of spaces used for indenting.
        lvl : int, optional
            Internal parameter for recursion, don't use!

        Returns
        -------
        s : str
            The formatted tree string.
        """
        space = indent * lvl * " "
        string = ""
        if self.type == self.PLAYLIST:
            string += space + f"Playlist: {self.name} ({self.entries} Tracks)\n"
        elif self.type == self.FOLDER:
            string += space + f"Folder: {self.name}\n"
            for node in self.get_playlists():
                string += node.treestr(indent, lvl + 1)
        return string

    def __eq__(self, other):
        return self.parent == other.parent and self.name == other.name

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.name})>"


# -- Main XML object -------------------------------------------------------------------


# noinspection PyPep8Naming,PyUnresolvedReferences
class RekordboxXml:
    """Rekordbox XML database object.

    The XML database contains the tracks and playlists in the Rekordbox collection. By
    importing the database, new tracks and items can be added to the Rekordbox
    collection.

    If a file path is passed to the constructor of the ``RekordboxXml`` object, the file
    is opened and parsed. Otherwise, an empty file is created with the given arguments.
    Creating an importable XML file requires a product name, xml database version and
    company name.

    Attributes
    ----------
    path : str, optional
        The file path to

    Examples
    --------
    Open Rekordbox XML file

    >>> file = RekordboxXml(Path(".testdata", "rekordbox 6", "database.xml"))

    Create new XML file

    >>> file = RekordboxXml()

    """

    ROOT_TAG = "DJ_PLAYLISTS"
    PRDT_TAG = "PRODUCT"
    PLST_TAG = "PLAYLISTS"
    COLL_TAG = "COLLECTION"

    def __init__(self, path=None, name=None, version=None, company=None):
        self._root = None
        self._product = None
        self._collection = None
        self._playlists = None
        self._root_node = None

        self._last_id = 0
        # Used for fast duplicate check
        self._locations = set()
        self._ids = set()

        if path is not None:
            self._parse(path)
        else:
            self._init(name, version, company)

    @property
    def frmt_version(self):
        """str : The version of the Rekordbox XML format."""
        return self._root.attrib["Version"]

    @property
    def product_name(self):
        """str : The product name that will be displayed in the software."""
        return self._product.attrib.get("Name")

    @property
    def product_version(self):
        """str : The product version."""
        return self._product.attrib.get("Version")

    @property
    def product_company(self):
        """str : The company name."""
        return self._product.attrib.get("Company")

    @property
    def num_tracks(self):
        """str : The number of tracks in the collection."""
        return int(self._collection.attrib.get("Entries"))

    @property
    def root_playlist_folder(self):
        """Node: The node of the root playlist folder containing all other nodes."""
        return self._root_node

    def _parse(self, path):
        """Parse an existing XML file.

        Parameters
        ----------
        path : str or Path
            The path to the XML file to parse.
        """
        tree = xml.parse(str(path))
        self._root = tree.getroot()
        self._product = self._root.find(self.PRDT_TAG)
        self._collection = self._root.find(self.COLL_TAG)
        self._playlists = self._root.find(self.PLST_TAG)
        self._root_node = Node(element=self._playlists.find(Node.TAG))
        self._update_cache()

    def _init(self, name=None, version=None, company=None, frmt_version=None):
        """Initialize a new XML file."""
        frmt_version = frmt_version or "1.0.0"
        name = name or "pyrekordbox"
        version = version or "0.0.1"
        company = company or ""

        # Initialize root element
        self._root = xml.Element(self.ROOT_TAG, attrib={"Version": frmt_version})
        # Initialize product element
        attrib = {"Name": name, "Version": version, "Company": company}
        self._product = xml.SubElement(self._root, self.PRDT_TAG, attrib=attrib)
        # Initialize collection element
        attrib = {"Entries": "0"}
        self._collection = xml.SubElement(self._root, self.COLL_TAG, attrib=attrib)
        # Initialize playlist element
        self._playlists = xml.SubElement(self._root, self.PLST_TAG)
        self._root_node = Node.folder(self._playlists, "ROOT")

        track_ids = self.get_track_ids()
        if track_ids:
            self._last_id = max(track_ids)

    def get_tracks(self):
        """Returns the tracks in the collection of the XML file.

        Returns
        -------
        tracks : list of Track
            A list of the track objects in the collection.
        """
        elements = self._collection.findall(f".//{Track.TAG}")
        return [Track(element=el) for el in elements]

    def get_track(self, index=None, TrackID=None, Location=None):
        """Get a track in the collection of the XML file.

        Parameters
        ----------
        index : int, optional
            If `index` is given, the track with this index in the collection is
            returned.
        TrackID : int, optional
            If `TrackID` is given, the track with this ID in the collection is
            returned.
        Location : str, optional
            If `Location` is given, the track with this file path in the collection is
            returned.

        Returns
        -------
        track : Track
            The XML track element.

        Raises
        ------
        ValueError:
            Raised if neither the index of the track id is specified.

        Examples
        --------
        Get track by index

        >>> file = RekordboxXml("database.xml")
        >>> track = file.get_track(0)

        or by ``TrackID``

        >>> track = file.get_track(TrackID=1)

        """
        if index is None and TrackID is None and Location is None:
            raise ValueError("Either index, TrackID or Location has to be specified!")

        if TrackID is not None:
            el = self._collection.find(f'.//{Track.TAG}[@TrackID="{TrackID}"]')
        elif Location is not None:
            encoded = encode_path(Location)
            el = self._collection.find(f'.//{Track.TAG}[@Location="{encoded}"]')
        else:
            el = self._collection.find(f".//{Track.TAG}[{index + 1}]")
        return Track(element=el)

    def get_track_ids(self):
        """Returns the `TrackID` of all tracks in the collection of the XML file.

        Returns
        -------
        ids : list of int
            The ID's of all tracks.
        """
        elements = self._collection.findall(f".//{Track.TAG}")
        return [int(el.attrib["TrackID"]) for el in elements]

    def get_playlist(self, *names):
        """Returns a playlist or playlist folder with the given path.

        Parameters
        ----------
        *names : str
            Names in the path. If no names are given the root playlist folder is
            returned.

        Returns
        -------
        node : Node
            The playlist or playlist folder node.

        Examples
        --------
        >>> file = RekordboxXml("database.xml")
        >>> playlist = file.get_playlist("Folder", "Sub Playlist")

        """
        node = self._root_node
        if not names:
            return node
        for name in names:
            node = node.get_playlist(name)
        return node

    # def _update_track_count(self):
    #     """Updates the track count element."""
    #     num_tracks = len(self._collection.findall(f".//{Track.TAG}"))
    #     self._collection.attrib["Entries"] = str(num_tracks)

    def _increment_track_count(self):
        """Increment the track count element."""
        old = int(self._collection.attrib["Entries"])
        self._collection.attrib["Entries"] = str(old + 1)

    def _decrement_track_count(self):
        """Decrement the track count element."""
        old = int(self._collection.attrib["Entries"])
        self._collection.attrib["Entries"] = str(old - 1)

    def _add_cache(self, track):
        """Add the TrackID and Location to the cache."""
        self._locations.add(track.Location)
        self._ids.add(track.TrackID)

    def _remove_cache(self, track):
        """Remove the TrackID and Location from the cache."""
        self._locations.remove(track.Location)
        self._ids.remove(track.TrackID)

    def _update_cache(self):
        """Update the cache with the current tracks in the collection."""
        self._locations.clear()
        self._ids.clear()
        for track in self.get_tracks():
            self._add_cache(track)

    def add_track(self, location, **kwargs):
        """Add a new track element to the Rekordbox XML collection.

        Parameters
        ----------
        location : str or Path
            The file path of the track.
        kwargs :
            Keyword arguments which are used to fill the track attributes. If no
            argument for ``TrackID`` is given the ID is auto-incremented.

        Returns
        -------
        track : Track
            The newly created XML track element.

        Raises
        ------
        ValueError:
            Raised if the database already contains a track with the track-id
            or file path.

        Examples
        --------
        >>> file = RekordboxXml("database.xml")
        >>> _ = file.add_track("path/to/track.wav")
        """
        if "TrackID" not in kwargs:
            kwargs["TrackID"] = self._last_id + 1

        # Check that Location and TrackID are unique
        track_id = kwargs["TrackID"]
        if os.path.normpath(location) in self._locations:
            raise XmlDuplicateError("Location", location)
        if track_id in self._ids:
            raise XmlDuplicateError("TrackID", track_id)

        # Create track and add it to the collection
        track = Track(self._collection, location, **kwargs)
        self._last_id = int(track["TrackID"])
        self._increment_track_count()
        self._add_cache(track)
        return track

    def remove_track(self, track):
        """Remove a track element from the Rekordbox XML collection.

        Parameters
        ----------
        track : Track
            The XML track element to remove.

        Examples
        --------
        >>> file = RekordboxXml("database.xml")
        >>> t = file.get_track(0)
        >>> file.remove_track(t)

        """
        self._collection.remove(track._element)  # noqa
        self._decrement_track_count()
        self._remove_cache(track)

    def add_playlist_folder(self, name):
        """Add a new top-level playlist folder to the XML collection.

        Parameters
        ----------
        name : str
            The name of the new playlist folder.

        Returns
        -------
        folder_node : Node
            The newly created playlist folder node.

        See Also
        --------
        Node.add_playlist_folder

        Examples
        --------
        >>> file = RekordboxXml("database.xml")
        >>> file.add_playlist_folder("New Folder")
        """
        return self._root_node.add_playlist_folder(name)

    def add_playlist(self, name, keytype="TrackID"):
        """Add a new top-level playlist to the XML collection.

        Parameters
        ----------
        name : str
            The name of the new playlist.
        keytype : {'TrackID', 'Location'} str
            The type of key the playlist uses to store the tracks. Can either be
            'TrackID' or 'Location'.

        Returns
        -------
        playlist_node : Node
            The newly created playlist node.

        See Also
        --------
        Node.add_playlist

        Examples
        --------
        Create playlist using the track ID as keys

        >>> file = RekordboxXml("database.xml")
        >>> file.add_playlist("New Playlist", keytype="TrackID")

        Create playlist using the file paths as keys

        >>> file.add_playlist("New Playlist 2", keytype="Location")
        """
        return self._root_node.add_playlist(name, keytype)

    def tostring(self, indent=None):
        """Returns the contents of the XML file as a string.

        Parameters
        ----------
        indent : str, optional
            The indentation used for formatting the XML file. The default is 3 spaces.

        Returns
        -------
        s : str
            The contents of the XML file
        """
        # Check track count is valid
        num_tracks = len(self._collection.findall(f".//{Track.TAG}"))
        n = int(self._collection.attrib["Entries"])
        if n != num_tracks:
            raise ValueError(f"Track count {num_tracks} does not match number of elements {n}")
        # Generate XML string
        return pretty_xml(self._root, indent, encoding="utf-8")

    def save(self, path="", indent=None):
        """Saves the contents to an XML file.

        Parameters
        ----------
        path : str or Path, optional
            The path for saving the XML file. The default is the original file.
        indent : str, optional
            The indentation used for formatting the XML element.
            The default is 3 spaces.
        """
        string = self.tostring(indent)
        with open(path, "w", encoding="utf-8") as fh:
            fh.write(string)

    def __repr__(self):
        name = self.product_name
        v = self.product_version
        company = self.product_company
        tracks = self.num_tracks

        cls = self.__class__.__name__
        s = f"{cls}(tracks={tracks}, info={name}, {company}, v{v})"
        return s



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

import pyrekordbox


# -- Project information -----------------------------------------------------

project = "pyrekordbox"
copyright = "2022-2023, Dylan Jones"
author = "Dylan L. Jones"
release = pyrekordbox.__version__
version = release
for sp in "abcfr":
    version = version.split(sp)[0]

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "4.4"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "numpydoc",
    "myst_parser",
    "numpydoc",
    "sphinx_copybutton",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "matplotlib.sphinxext.plot_directive",
    "sphinx.ext.intersphinx",  # links to numpy, scipy ... docs
    "sphinx.ext.coverage",
    "sphinx.ext.extlinks",  # define roles for links
    "sphinx.ext.viewcode",
]

# If you need extensions of a certain version or higher, list them here.
needs_extensions = {"myst_parser": "0.13.7"}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "tests"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = [".rst", ".md"]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_theme_options = {
    "light_logo": "logos/dark/logo_primary.svg",
    "dark_logo": "logos/light/logo_primary.svg",
    "sidebar_hide_name": True,
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/dylanljones/pyrekordbox",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
}

# html_title = f"{release} documentation"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = "sphinx"
# pygments_dark_style = "monokai"

# We need headers to be linkable to so ask MyST-Parser to autogenerate anchor IDs for
# headers up to and including level 3.
myst_heading_anchors = 3

# Prettier support formatting some MyST syntax but not all, so let's disable the
# unsupported yet still enabled by default ones.
myst_disable_syntax = [
    "colon_fence",
    "myst_block_break",
    "myst_line_comment",
    "math_block",
]


# Don't show type hints
autodoc_typehints = "none"

# Preserve order
autodoc_member_order = "bysource"


# -- Apidoc ------------------------------------------------------------------

add_module_names = True


# -- Autosummary -------------------------------------------------------------

autosummary_generate = True
# autosummary_imported_members = True


# -- Numpy extension ---------------------------------------------------------

numpydoc_use_plots = True
# numpydoc_xref_param_type = True
# numpydoc_xref_ignore = "all"  # not working...
numpydoc_show_class_members = False


# -- Intersphinx -------------------------------------------------------------

# taken from https://gist.github.com/bskinn/0e164963428d4b51017cebdb6cda5209
intersphinx_mapping = {
    "python": (r"https://docs.python.org", None),
    "numpy": (r"https://docs.scipy.org/doc/numpy/", None),
    "np": (r"https://docs.scipy.org/doc/numpy/", None),
    "matplotlib": (r"https://matplotlib.org/", None),
    "<name>": ("https://docs.python.org/3/", None),
}


# -- Auto-run sphinx-apidoc --------------------------------------------------


def run_apidoc(_):
    from sphinx.ext.apidoc import main
    import os
    import sys

    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    proj_dir = os.path.dirname(os.path.dirname(cur_dir))
    doc_dir = os.path.join(proj_dir, "docs")
    output_path = os.path.join(doc_dir, "source", "generated")
    module = os.path.join(proj_dir, "pyrekordbox")
    exclude = os.path.join(module, "tests")
    template_dir = os.path.join(doc_dir, "source", "_templates", "apidoc")
    main(["-fMeT", "-o", output_path, module, exclude, "--templatedir", template_dir])


def setup(app):
    app.connect("builder-inited", run_apidoc)



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2022-05-07

import os

import pytest

from pyrekordbox import RekordboxXml
from pyrekordbox.rbxml import Node, PositionMark, Tempo, XmlDuplicateError

TEST_ROOT = os.path.join(os.path.dirname(os.path.dirname(__file__)), ".testdata")
XML5 = os.path.join(TEST_ROOT, "rekordbox 5", "database.xml")
XML6 = os.path.join(TEST_ROOT, "rekordbox 6", "database.xml")


V5_TRACK_INFOS = [
    {
        "TrackID": 1,
        "Name": "NOISE",
        "Artist": "",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "WAV File",
        "Size": 1382226,
        "TotalTime": 5,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 0.0,
        "DateAdded": "2022-04-04",
        "BitRate": 2116,
        "SampleRate": 44100.0,
        "Comments": "",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Sampler/OSC_SAMPLER/PRESET ONESHOT/NOISE.wav",
        "Remixer": "",
        "Tonality": "",
        "Label": "",
        "Mix": "",
    },
    {
        "TrackID": 2,
        "Name": "SINEWAVE",
        "Artist": "",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "WAV File",
        "Size": 1515258,
        "TotalTime": 5,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 0.0,
        "DateAdded": "2022-04-04",
        "BitRate": 2116,
        "SampleRate": 44100.0,
        "Comments": "",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Sampler/OSC_SAMPLER/PRESET ONESHOT/SINEWAVE.wav",  # noqa: E501
        "Remixer": "",
        "Tonality": "",
        "Label": "",
        "Mix": "",
    },
    {
        "TrackID": 3,
        "Name": "SIREN",
        "Artist": "",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "WAV File",
        "Size": 1941204,
        "TotalTime": 7,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 0.0,
        "DateAdded": "2022-04-04",
        "BitRate": 2116,
        "SampleRate": 44100.0,
        "Comments": "",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Sampler/OSC_SAMPLER/PRESET ONESHOT/SIREN.wav",
        "Remixer": "",
        "Tonality": "",
        "Label": "",
        "Mix": "",
    },
    {
        "TrackID": 4,
        "Name": "HORN",
        "Artist": "",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "WAV File",
        "Size": 2010816,
        "TotalTime": 7,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 0.0,
        "DateAdded": "2022-04-04",
        "BitRate": 2116,
        "SampleRate": 44100.0,
        "Comments": "",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Sampler/OSC_SAMPLER/PRESET ONESHOT/HORN.wav",
        "Remixer": "",
        "Tonality": "",
        "Label": "",
        "Mix": "",
    },
    {
        "TrackID": 5,
        "Name": "Demo Track 1",
        "Artist": "Loopmasters",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "MP3 File",
        "Size": 6899624,
        "TotalTime": 172,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 128.0,
        "DateAdded": "2022-04-04",
        "BitRate": 320,
        "SampleRate": 44100.0,
        "Comments": "Tracks by www.loopmasters.com",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Demo Tracks/Demo Track 1.mp3",
        "Remixer": "",
        "Tonality": "",
        "Label": "Loopmasters",
        "Mix": "",
    },
    {
        "TrackID": 6,
        "Name": "Demo Track 2",
        "Artist": "Loopmasters",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "MP3 File",
        "Size": 5124342,
        "TotalTime": 128,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 120.0,
        "DateAdded": "2022-04-04",
        "BitRate": 320,
        "SampleRate": 44100.0,
        "Comments": "Tracks by www.loopmasters.com",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Demo Tracks/Demo Track 2.mp3",
        "Remixer": "",
        "Tonality": "",
        "Label": "Loopmasters",
        "Mix": "",
    },
]

V6_TRACK_INFOS = [
    {
        "TrackID": 253529738,
        "Name": "Demo Track 1",
        "Artist": "Loopmasters",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "Mp3-Datei ",
        "Size": 6899624,
        "TotalTime": 172,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 128.0,
        "DateAdded": "2022-04-09",
        "BitRate": 320,
        "SampleRate": 44100.0,
        "Comments": "Tracks by www.loopmasters.com",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Demo Tracks/Demo Track 1.mp3",
        "Remixer": "",
        "Tonality": "Fm",
        "Label": "Loopmasters",
        "Mix": "",
    },
    {
        "TrackID": 17109519,
        "Name": "Demo Track 2",
        "Artist": "Loopmasters",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "Mp3-Datei ",
        "Size": 5124342,
        "TotalTime": 128,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 120.0,
        "DateAdded": "2022-04-09",
        "BitRate": 320,
        "SampleRate": 44100.0,
        "Comments": "Tracks by www.loopmasters.com",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Demo Tracks/Demo Track 2.mp3",
        "Remixer": "",
        "Tonality": "Fm",
        "Label": "Loopmasters",
        "Mix": "",
    },
    {
        "TrackID": 49557014,
        "Name": "HORN",
        "Artist": "",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "Wav-Datei ",
        "Size": 2010816,
        "TotalTime": 7,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 0.0,
        "DateAdded": "2022-04-09",
        "BitRate": 2116,
        "SampleRate": 44100.0,
        "Comments": "",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Sampler/OSC_SAMPLER/PRESET ONESHOT/HORN.wav",
        "Remixer": "",
        "Tonality": "",
        "Label": "",
        "Mix": "",
    },
    {
        "TrackID": 209873516,
        "Name": "NOISE",
        "Artist": "",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "Wav-Datei ",
        "Size": 1382226,
        "TotalTime": 5,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 0.0,
        "DateAdded": "2022-04-09",
        "BitRate": 2116,
        "SampleRate": 44100.0,
        "Comments": "",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Sampler/OSC_SAMPLER/PRESET ONESHOT/NOISE.wav",
        "Remixer": "",
        "Tonality": "",
        "Label": "",
        "Mix": "",
    },
    {
        "TrackID": 55231398,
        "Name": "SINEWAVE",
        "Artist": "",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "Wav-Datei ",
        "Size": 1515258,
        "TotalTime": 5,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 0.0,
        "DateAdded": "2022-04-09",
        "BitRate": 2116,
        "SampleRate": 44100.0,
        "Comments": "",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Sampler/OSC_SAMPLER/PRESET ONESHOT/SINEWAVE.wav",  # noqa: E501
        "Remixer": "",
        "Tonality": "",
        "Label": "",
        "Mix": "",
    },
    {
        "TrackID": 92396897,
        "Name": "SIREN",
        "Artist": "",
        "Composer": "",
        "Album": "",
        "Grouping": "",
        "Genre": "",
        "Kind": "Wav-Datei ",
        "Size": 1941204,
        "TotalTime": 7,
        "DiscNumber": 0,
        "TrackNumber": 0,
        "Year": 0,
        "AverageBpm": 0.0,
        "DateAdded": "2022-04-09",
        "BitRate": 2116,
        "SampleRate": 44100.0,
        "Comments": "",
        "PlayCount": 0,
        "Rating": 0,
        "Location": "C:/Music/PioneerDJ/Sampler/OSC_SAMPLER/PRESET ONESHOT/SIREN.wav",
        "Remixer": "",
        "Tonality": "",
        "Label": "",
        "Mix": "",
    },
]

TEMPOS = [
    [{"Inizio": 0.025, "Bpm": 128.0, "Metro": "4/4", "Battito": 1}],
    [
        {"Inizio": 0.025, "Bpm": 120.0, "Metro": "4/4", "Battito": 1},
        {"Inizio": 48.026, "Bpm": 120.0, "Metro": "4/4", "Battito": 1},
        {"Inizio": 48.525, "Bpm": 120.0, "Metro": "4/4", "Battito": 2},
        {"Inizio": 49.026, "Bpm": 120.0, "Metro": "4/4", "Battito": 3},
        {"Inizio": 49.525, "Bpm": 120.0, "Metro": "4/4", "Battito": 4},
        {"Inizio": 50.026, "Bpm": 120.0, "Metro": "4/4", "Battito": 1},
        {"Inizio": 50.525, "Bpm": 120.0, "Metro": "4/4", "Battito": 2},
        {"Inizio": 51.026, "Bpm": 120.0, "Metro": "4/4", "Battito": 3},
        {"Inizio": 51.525, "Bpm": 120.0, "Metro": "4/4", "Battito": 4},
        {"Inizio": 52.026, "Bpm": 120.0, "Metro": "4/4", "Battito": 1},
    ],
]

POSITION_MARKS = [
    [
        {"Name": "", "Type": "cue", "Start": 0.025, "Num": -1},
        {"Name": "", "Type": "cue", "Start": 15.025, "Num": -1},
        {"Name": "", "Type": "cue", "Start": 30.025, "Num": -1},
        {"Name": "", "Type": "cue", "Start": 45.025, "Num": -1},
    ],
    [
        {"Name": "", "Type": "cue", "Start": 0.025, "Num": -1},
        {"Name": "", "Type": "cue", "Start": 16.025, "Num": -1},
        {"Name": "", "Type": "cue", "Start": 32.025, "Num": -1},
        {"Name": "", "Type": "cue", "Start": 48.026, "Num": -1},
    ],
]


def test_track_attribs_v5():
    xml = RekordboxXml(XML5)

    for i, track in enumerate(xml.get_tracks()):
        for attr in track.ATTRIBS:
            info = V5_TRACK_INFOS[i]
            value = track[attr]
            if attr == "Location":
                assert os.path.normpath(value) == os.path.normpath(info[attr])
            elif attr in info:
                assert value == info[attr]
            else:
                assert value is None


def test_track_attribs_v6():
    xml = RekordboxXml(XML6)

    for i, track in enumerate(xml.get_tracks()):
        for attr in track.ATTRIBS:
            info = V6_TRACK_INFOS[i]
            value = track[attr]
            if attr == "Location":
                assert os.path.normpath(value) == os.path.normpath(info[attr])
            elif attr in info:
                assert value == info[attr]
            else:
                assert value is None


def test_tempos_v5():
    xml = RekordboxXml(XML5)

    for i in range(3):
        track = xml.get_track(i)
        assert len(track.tempos) == 0

    track = xml.get_track(4)
    tempos = TEMPOS[0]
    for info, tempo in zip(tempos, track.tempos):
        for attr in Tempo.ATTRIBS:
            value = tempo[attr]
            if attr in info:
                assert value == info[attr]
            else:
                assert value is None

    track = xml.get_track(5)
    tempos = TEMPOS[1]
    for info, tempo in zip(tempos, track.tempos):
        for attr in Tempo.ATTRIBS:
            value = tempo[attr]
            if attr in info:
                assert value == info[attr]
            else:
                assert value is None


def test_marks_v5():
    xml = RekordboxXml(XML5)

    for i in range(3):
        track = xml.get_track(i)
        assert len(track.tempos) == 0

    track = xml.get_track(4)
    tempos = POSITION_MARKS[0]
    for info, posmark in zip(tempos, track.marks):
        for attr in PositionMark.ATTRIBS:
            value = posmark[attr]
            if attr in info:
                assert value == info[attr]
            else:
                assert value is None

    track = xml.get_track(5)
    tempos = POSITION_MARKS[1]
    for info, posmark in zip(tempos, track.marks):
        for attr in PositionMark.ATTRIBS:
            value = posmark[attr]
            if attr in info:
                assert value == info[attr]
            else:
                assert value is None


def test_add_track():
    xml = RekordboxXml()
    track1 = xml.add_track("C:/path/to/file1.wav")

    raw_location = os.path.normpath(track1._element.attrib["Location"])
    assert track1.TrackID == 1
    assert raw_location == os.path.normpath("file:/localhost/C:/path/to/file1.wav")
    assert track1.Location == os.path.normpath("C:/path/to/file1.wav")

    # test auto-increment of TrackID
    track2 = xml.add_track("C:/path/to/file2.wav")
    assert track2.TrackID == 2

    # test manual TrackID
    track3 = xml.add_track("C:/path/to/file3.wav", TrackID=10)
    assert track3.TrackID == 10

    # test auto-increment after manual TrackID
    track4 = xml.add_track("C:/path/to/file4.wav")
    assert track4.TrackID == 11

    # Location exists
    with pytest.raises(XmlDuplicateError):
        xml.add_track("C:/path/to/file1.wav")

    # TrackID exists
    with pytest.raises(XmlDuplicateError):
        xml.add_track("C:/path/to/file_new.wav", TrackID=track1.TrackID)


def test_update_track_count():
    xml = RekordboxXml()
    track1 = xml.add_track("C:/path/to/file1.wav")
    assert xml.num_tracks == 1

    track2 = xml.add_track("C:/path/to/file2.wav")
    assert xml.num_tracks == 2

    xml.remove_track(track1)
    assert xml.num_tracks == 1

    xml.remove_track(track2)
    assert xml.num_tracks == 0


def test_add_tempo():
    xml = RekordboxXml()
    track = xml.add_track("C:/path/to/file1.wav")
    assert len(track.tempos) == 0

    inizio = 0.0
    bpm = 100.0
    metro = "4/4"
    battito = 1
    track.add_tempo(inizio, bpm, metro, battito)

    assert len(track.tempos) == 1
    tempo = track.tempos[0]
    assert tempo.Inizio == inizio
    assert tempo.Bpm == bpm
    assert tempo.Metro == metro
    assert tempo.Battito == battito


def test_add_position_mark():
    xml = RekordboxXml()
    track = xml.add_track("C:/path/to/file1.wav")
    assert len(track.marks) == 0

    name = "cue point"
    mark_type = "cue"
    start = 0.0
    end = None
    num = -1
    track.add_mark(name, mark_type, start, end, num)

    assert len(track.marks) == 1

    mark = track.marks[0]
    assert mark.Name == name
    assert mark.Type == mark_type
    assert mark.Start == start
    assert mark.End == end
    assert mark.Num == num


def test_get_track():
    xml = RekordboxXml(XML6)

    with pytest.raises(ValueError):
        # Nothing provided
        xml.get_track()

    idx = 0
    expected_loc = os.path.normpath(r"C:/Music/PioneerDJ/Demo Tracks/Demo Track 1.mp3")
    expected_id = 253529738

    # Test getting by index
    track = xml.get_track(index=idx)
    assert track.Location == expected_loc
    assert track.TrackID == expected_id

    # Test getting by TrackID
    track = xml.get_track(TrackID=expected_id)
    assert track.Location == expected_loc
    assert track.TrackID == expected_id

    # Test getting by Location
    track = xml.get_track(Location=expected_loc)
    assert track.Location == expected_loc
    assert track.TrackID == expected_id


def test_get_playlist():
    xml = RekordboxXml(XML5)

    playlists = xml.get_playlist()
    assert isinstance(playlists, Node)

    playlist = xml.get_playlist("Playlist1")
    assert playlist.is_playlist
    assert not playlist.is_folder

    folder = xml.get_playlist("Folder")
    assert folder.is_folder
    assert not folder.is_playlist

    sub_playlists = folder.get_playlists()
    assert isinstance(sub_playlists, list)
    assert len(sub_playlists) == 1

    sub_playlist = xml.get_playlist("Folder", "Sub Playlist")
    assert sub_playlist.is_playlist
    assert not sub_playlist.is_folder

    assert folder.get_playlist("Sub Playlist").name == sub_playlist.name


def test_playlist_entries():
    xml = RekordboxXml(XML5)

    playlist = xml.get_playlist("Playlist1")
    key_type = playlist.key_type
    for key in playlist.get_tracks():
        kwargs = {key_type: key}
        xml.get_track(**kwargs)


def test_add_playlist():
    xml = RekordboxXml()
    playlist = xml.add_playlist("Playlist")

    assert playlist.is_playlist
    assert not playlist.is_folder

    with pytest.raises(ValueError):
        playlist.add_playlist("Sub Playlist")


def test_add_playlist_folder():
    xml = RekordboxXml()
    folder = xml.add_playlist_folder("Folder")

    assert not folder.is_playlist
    assert folder.is_folder

    folder.add_playlist("Sub Playlist")


def test_update_folder_count():
    xml = RekordboxXml()
    folder = xml.add_playlist_folder("Folder")

    folder.add_playlist("P1")
    assert folder.count == 1

    folder.add_playlist_folder("F1")
    assert folder.count == 2

    folder.remove_playlist("P1")
    assert folder.count == 1

    folder.remove_playlist("F1")
    assert folder.count == 0


def test_update_playlist_entries():
    xml = RekordboxXml()
    playlist = xml.add_playlist("Playlist")

    playlist.add_track(0)
    assert playlist.entries == 1
    assert playlist.get_track(0) == 0

    playlist.add_track(1)
    assert playlist.entries == 2

    playlist.remove_track(1)
    assert playlist.entries == 1

    playlist.remove_track(0)
    assert playlist.entries == 0


def test_playlist_treestr():
    xml = RekordboxXml(XML5)
    root = xml.get_playlist()
    assert isinstance(root.treestr(), str)

    playlist = xml.get_playlist("Playlist1")
    assert isinstance(playlist.treestr(), str)


def test_save_load(tmp_path):
    xml = RekordboxXml(XML5)

    # Save to a file
    test_file = tmp_path / "test_save_reload.xml"
    xml.save(test_file)

    # Load the saved file
    loaded = RekordboxXml(test_file)

    # Verify content was preserved
    assert loaded.num_tracks == xml.num_tracks



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-12-13

import logging
import xml.etree.cElementTree as xml
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, IntEnum
from typing import List, Union

from dateutil.relativedelta import relativedelta  # noqa
from sqlalchemy import and_, not_, or_
from sqlalchemy.sql.elements import BooleanClauseList

from .tables import DjmdContent

logger = logging.getLogger(__name__)

__all__ = [
    "LogicalOperator",
    "Property",
    "Operator",
    "Condition",
    "SmartList",
]


class LogicalOperator(IntEnum):
    ALL = 1
    ANY = 2


class Operator(IntEnum):
    EQUAL = 1
    NOT_EQUAL = 2
    GREATER = 3
    LESS = 4
    IN_RANGE = 5
    IN_LAST = 6
    NOT_IN_LAST = 7
    CONTAINS = 8
    NOT_CONTAINS = 9
    STARTS_WITH = 10
    ENDS_WITH = 11


class Property(str, Enum):
    ARTIST = "artist"
    ALBUM = "album"
    ALBUM_ARTIST = "albumArtist"
    ORIGINAL_ARTIST = "originalArtist"
    BPM = "bpm"
    GROUPING = "grouping"
    COMMENTS = "comments"
    PRODUCER = "producer"
    STOCK_DATE = "stockDate"
    DATE_CREATED = "dateCreated"
    COUNTER = "counter"
    FILENAME = "fileName"
    GENRE = "genre"
    KEY = "key"
    LABEL = "label"
    MIX_NAME = "mixName"
    MYTAG = "myTag"
    RATING = "rating"
    DATE_RELEASED = "dateReleased"
    REMIXED_BY = "remixedBy"
    DURATION = "duration"
    NAME = "name"
    YEAR = "year"


_STR_OPS = [
    Operator.EQUAL,
    Operator.NOT_EQUAL,
    Operator.CONTAINS,
    Operator.NOT_CONTAINS,
    Operator.STARTS_WITH,
    Operator.ENDS_WITH,
]

_NUM_OPS = [
    Operator.EQUAL,
    Operator.NOT_EQUAL,
    Operator.GREATER,
    Operator.LESS,
    Operator.IN_RANGE,
]

_DATE_OPS = [
    Operator.EQUAL,
    Operator.NOT_EQUAL,
    Operator.GREATER,
    Operator.LESS,
    Operator.IN_RANGE,
    Operator.IN_LAST,
    Operator.NOT_IN_LAST,
]

# Defines the valid operators for each property
VALID_OPS = {
    Property.ARTIST: _STR_OPS,
    Property.ALBUM: _STR_OPS,
    Property.ALBUM_ARTIST: _STR_OPS,
    Property.ORIGINAL_ARTIST: _STR_OPS,
    Property.BPM: _NUM_OPS,
    Property.GROUPING: [Operator.EQUAL, Operator.NOT_EQUAL],
    Property.COMMENTS: _STR_OPS,
    Property.PRODUCER: _STR_OPS,
    Property.STOCK_DATE: _DATE_OPS,
    Property.DATE_CREATED: _DATE_OPS,
    Property.COUNTER: _NUM_OPS,
    Property.FILENAME: _STR_OPS,
    Property.GENRE: _STR_OPS,
    Property.KEY: _STR_OPS,
    Property.LABEL: _STR_OPS,
    Property.MIX_NAME: _STR_OPS,
    Property.MYTAG: [Operator.CONTAINS, Operator.NOT_CONTAINS],
    Property.RATING: _NUM_OPS,
    Property.DATE_RELEASED: _DATE_OPS,
    Property.REMIXED_BY: _STR_OPS,
    Property.DURATION: _NUM_OPS,
    Property.NAME: _STR_OPS,
    Property.YEAR: _NUM_OPS,
}

# Defines the column names in the DB for properties that are directly mapped
PROPERTY_COLUMN_MAP = {
    Property.ARTIST: "ArtistName",
    Property.ALBUM: "AlbumName",
    Property.ALBUM_ARTIST: "AlbumArtistName",
    Property.ORIGINAL_ARTIST: "OrgArtistName",
    Property.BPM: "BPM",
    Property.GROUPING: "ColorID",
    Property.COMMENTS: "Commnt",
    Property.PRODUCER: "ComposerName",
    Property.STOCK_DATE: "StockDate",
    Property.DATE_CREATED: "created_at",
    Property.COUNTER: "DJPlayCount",
    Property.FILENAME: "FileNameL",
    Property.GENRE: "GenreName",
    Property.KEY: "KeyName",
    Property.LABEL: "LabelName",
    # Property.MIX_NAME don't know what this maps to
    Property.MYTAG: "MyTagIDs",
    Property.RATING: "Rating",
    Property.DATE_RELEASED: "ReleaseDate",
    Property.REMIXED_BY: "RemixerName",
    Property.DURATION: "Length",
    Property.NAME: "Title",
    Property.YEAR: "ReleaseYear",
}

TYPE_CONVERSION = {
    Property.BPM: int,
    Property.STOCK_DATE: lambda x: datetime.strptime(x, "%Y-%m-%d"),
    Property.DATE_CREATED: lambda x: datetime.strptime(x, "%Y-%m-%d"),
    Property.COUNTER: int,
    Property.RATING: int,
    Property.DATE_RELEASED: lambda x: datetime.strptime(x, "%Y-%m-%d"),
    Property.DURATION: int,
    Property.YEAR: int,
}

PROPERTIES = [str(p.value) for p in list(Property)]  # noqa


@dataclass
class Condition:
    """Dataclass for a smart playlist condition."""

    property: str
    operator: int
    unit: str
    value_left: Union[str, int]
    value_right: Union[str, int]

    def __post_init__(self):
        if self.property not in PROPERTIES:
            raise ValueError(
                f"Invalid property: '{self.property}'! Supported properties: {PROPERTIES}"
            )

        valid_ops = VALID_OPS[self.property]
        if self.operator not in valid_ops:
            raise ValueError(
                f"Invalid operator '{self.operator}' for '{self.property}', "
                f"must be one of {valid_ops}"
            )

        if self.operator == Operator.IN_RANGE:
            if not self.value_right:
                raise ValueError(f"Operator '{self.operator}' requires `value_right`")


def left_bitshift(x: int, nbit: int = 32) -> int:
    """Left shifts an N bit integer with sign change."""
    return x - 2**nbit


def right_bitshift(x: int, nbit: int = 32) -> int:
    """Right shifts an N bit integer with sign change."""
    return x + 2**nbit


def _get_condition_values(cond):
    val_left = cond.value_left
    val_right = cond.value_right
    func = None
    if cond.operator in (Operator.IN_LAST, Operator.NOT_IN_LAST):
        func = int
    elif cond.property in TYPE_CONVERSION:
        func = TYPE_CONVERSION[cond.property]

    if func is not None:
        if val_left != "":
            val_left = func(val_left)
        if val_right != "":
            try:
                val_right = func(val_right)
            except ValueError:
                pass

    if val_left == "":
        val_left = None

    return val_left, val_right


class SmartList:
    """Rekordbox smart playlist XML handler."""

    def __init__(self, logical_operator: int = LogicalOperator.ALL, auto_update: int = 0):
        self.playlist_id: Union[int, str] = ""
        self.logical_operator: int = int(logical_operator)
        self.auto_update: int = auto_update
        self.conditions: List[Condition] = list()

    def parse(self, source: str) -> None:
        """Parse the XML source of a smart playlist."""
        tree = xml.ElementTree(xml.fromstring(source))
        root = tree.getroot()
        conditions = list()
        for child in root.findall("CONDITION"):
            condition = Condition(
                property=child.attrib["PropertyName"],
                operator=int(child.attrib["Operator"]),
                unit=child.attrib["ValueUnit"],
                value_left=child.attrib["ValueLeft"],
                value_right=child.attrib["ValueRight"],
            )
            conditions.append(condition)

        self.playlist_id = str(right_bitshift(int(root.attrib["Id"])))
        self.logical_operator = int(root.attrib["LogicalOperator"])
        self.auto_update = int(root.attrib["AutomaticUpdate"])
        self.conditions = conditions

    def to_xml(self) -> str:
        """Convert the smart playlist conditions to XML."""
        attrib = {
            "Id": str(left_bitshift(int(self.playlist_id))),
            "LogicalOperator": str(self.logical_operator),
            "AutomaticUpdate": str(self.auto_update),
        }
        root = xml.Element("NODE", attrib=attrib)
        for cond in self.conditions:
            attrib = {
                "PropertyName": str(cond.property),
                "Operator": str(cond.operator),
                "ValueUnit": str(cond.unit),
                "ValueLeft": str(cond.value_left),
                "ValueRight": str(cond.value_right),
            }
            xml.SubElement(root, "CONDITION", attrib=attrib)
        return xml.tostring(root).decode("utf-8").replace(" /", "/")

    def add_condition(
        self,
        prop: str,
        operator: int,
        value_left: str,
        value_right: str = "",
        unit: str = "",
    ) -> None:
        """Add a condition to the smart playlist.

        Parameters
        ----------
        prop : str
            The property to filter on.
        operator : {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11} int
            The operator to use. Must be in the range of 1-11
        value_left : str
            The left value to use.
        value_right : str, optional
            The right value to use, by default "".
        unit : str, optional
            The unit to use, by default "".
        """
        if isinstance(prop, Property):
            prop = str(prop.value)
        cond = Condition(prop, int(operator), unit, value_left, value_right)
        self.conditions.append(cond)

    def filter_clause(self) -> BooleanClauseList:
        """Return a SQLAlchemy filter clause matching the content of the smart playlist.

        Returns
        -------
        BooleanClauseList
            A filter list macthing the contents of the smart playlist.
        """
        logical_op = and_ if self.logical_operator == LogicalOperator.ALL else or_

        comps = list()
        for cond in self.conditions:
            val_left, val_right = _get_condition_values(cond)
            # val_left = str(-abs(int(val_left))) if val_left is not None else ""
            if cond.property in PROPERTY_COLUMN_MAP:
                colum_name = PROPERTY_COLUMN_MAP[cond.property]
                if cond.property == Property.MYTAG:
                    if int(val_left) < 0:
                        val_left = str(right_bitshift(int(val_left)))

                if cond.operator == Operator.EQUAL:
                    comp = getattr(DjmdContent, colum_name) == val_left
                elif cond.operator == Operator.NOT_EQUAL:
                    comp = getattr(DjmdContent, colum_name) != val_left
                elif cond.operator == Operator.GREATER:
                    comp = getattr(DjmdContent, colum_name) > val_left
                elif cond.operator == Operator.LESS:
                    comp = getattr(DjmdContent, colum_name) < val_left
                elif cond.operator == Operator.IN_RANGE:
                    comp = getattr(DjmdContent, colum_name).between(val_left, val_right)
                elif cond.operator == Operator.CONTAINS:
                    comp = getattr(DjmdContent, colum_name).contains(val_left)
                elif cond.operator == Operator.NOT_CONTAINS:
                    comp = not_(getattr(DjmdContent, colum_name).contains(val_left))
                elif cond.operator == Operator.STARTS_WITH:
                    comp = getattr(DjmdContent, colum_name).startswith(val_left)
                elif cond.operator == Operator.ENDS_WITH:
                    comp = getattr(DjmdContent, colum_name).endswith(val_left)
                elif cond.operator == Operator.IN_LAST:
                    now = datetime.now()
                    if cond.unit == "day":
                        t0 = now - relativedelta(days=val_left)
                        comp = getattr(DjmdContent, colum_name) > t0
                    elif cond.unit == "month":
                        t0 = now - relativedelta(months=val_left)
                        comp = getattr(DjmdContent, colum_name).month > t0
                    else:
                        raise ValueError(f"Unknown unit '{cond.unit}'")
                elif cond.operator == Operator.NOT_IN_LAST:
                    now = datetime.now()
                    if cond.unit == "day":
                        t0 = now - relativedelta(days=val_left)
                        comp = getattr(DjmdContent, colum_name) < t0
                    elif cond.unit == "month":
                        t0 = now - relativedelta(months=val_left)
                        comp = getattr(DjmdContent, colum_name).month < t0
                    else:
                        raise ValueError(f"Unknown unit '{cond.unit}'")
                else:
                    raise ValueError(f"Unknown operator '{cond.operator}'")
                comps.append(comp)

            else:
                logger.warning(f"Unsupported property '{cond.property}'")

        return logical_op(*comps)



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2025-04-21

from subprocess import run
from types import SimpleNamespace as Namespace


def shell(command, **kwargs):
    """Execute a shell command capturing output and exit code."""
    completed = run(command, shell=True, capture_output=True, check=False, **kwargs)
    return Namespace(
        exit_code=completed.returncode,
        stdout=completed.stdout.decode(),
        stderr=completed.stderr.decode(),
    )


def test_cli():
    """Check if the CLI is callable."""
    result = shell("python -m pyrekordbox --help")
    # Check if the command was successful
    assert result.exit_code == 0, f"Command failed with exit code {result.exit_code}"


def test_download_key():
    """Test the download-key command."""
    result = shell("python -m pyrekordbox download-key")
    # Check if the command was successful
    assert result.exit_code == 0, f"Command failed with exit code {result.exit_code}"
    assert "Found key, updating cache file" in result.stdout



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2022-04-10

"""This module contains common constants and methods used in other modules."""

import os
import warnings
import xml.etree.cElementTree as xml
from xml.dom import minidom

import psutil

warnings.simplefilter("always", DeprecationWarning)


def warn_deprecated(name, new_name="", hint="", remove_in=""):
    s = f"'{name}' is deprecated"
    if remove_in:
        s += f" and will be removed in version '{remove_in}'"

    if new_name:
        s += f", use '{new_name}' instead!"
    else:
        s += "!"

    if hint:
        s += "\n" + hint

    warnings.warn(s, DeprecationWarning, stacklevel=3)


def get_process_id(name: str, raise_exec=False) -> int:
    """Returns the ID of a process if it exists.

    Parameters
    ----------
    name : str
        The name of the process, for example 'rekordbox'.
    raise_exec : bool, optional
        Raise an exception if the process can not be found.

    Returns
    -------
    pid : int
        The ID of the process if it exists, otherwise zero.

    Raises
    ------
    RuntimeError: If ``raise_exec=True``, raises a runtime error if the application
        is not running.

    Examples
    --------
    >>> get_process_id("rekordbox")
    12345

    >>> get_process_id("rekordboxAgent")
    23456
    """
    for proc in psutil.process_iter():
        try:
            proc_name = os.path.splitext(proc.name())[0]  # needed on Windows (.exe)
            if proc_name == name:
                return proc.pid
        except (psutil.AccessDenied, psutil.NoSuchProcess):
            pass
    if raise_exec:
        raise RuntimeError("No process with name 'rekordbox' found!")
    return 0


def get_rekordbox_pid(raise_exec=False):
    """Returns the process ID of the Rekordbox application.

    Parameters
    ----------
    raise_exec : bool, optional
        Raise an exception if the Rekordbox process can not be found.

    Returns
    -------
    pid : int
        The ID of the Rekordbox process if it exists, otherwise zero.

    Raises
    ------
    RuntimeError: If ``raise_exec=True``, raises a runtime error if the Rekordbox
        application is not running.

    Examples
    --------
    >>> get_rekordbox_pid()
    12345
    """
    return get_process_id("rekordbox", raise_exec)


def get_rekordbox_agent_pid(raise_exec=False):
    """Returns the process ID of the RekordboxAgent application.

    Parameters
    ----------
    raise_exec : bool, optional
        Raise an exception if the RekordboxAgent process can not be found.

    Returns
    -------
    pid : int
        The ID of the RekordboxAgent process if it exists, otherwise zero.

    Raises
    ------
    RuntimeError: If ``raise_exec=True``, raises a runtime error if the RekordboxAgent
        application is not running.

    Examples
    --------
    >>> get_rekordbox_agent_pid()
    23456
    """
    return get_process_id("rekordboxAgent", raise_exec)


def pretty_xml(element, indent=None, encoding="utf-8"):
    r"""Generates a formatted string of an XML element.

    Parameters
    ----------
    element : xml.etree.cElementTree.Element
        The input XML element.
    indent : str, optional
        The indentation used for formatting the XML element. The default is 3 spaces.
    encoding : str, optional
        The encoding used for formatting the XML element. The default is 'utf-8'.

    Returns
    -------
    xml_string : str
        The formatted string of the XML element.

    Notes
    -----
    This method is needed for Python 3.8 and below. Starting with Python 3.9 the XML
    module has a built-in pretty-print function:

    >>> tree = xml.ElementTree(root)  # noqa
    >>> xml.indent(tree, space="\t", level=0)
    >>> tree.write(path, encoding="utf-8", xml_declaration=True)  # noqa

    This method will be used as soon support for Python 3.8 and below is dropped.
    """
    # Build pretty xml-string
    if indent is None:
        indent = "    "
    if encoding is None:
        encoding = "utf-8"
    rough_string = xml.tostring(element, encoding)
    reparsed = minidom.parseString(rough_string)
    string = reparsed.toprettyxml(indent=indent, encoding=encoding).decode()
    # Remove annoying empty lines
    string = "\n".join([line for line in string.splitlines() if line.strip()])
    return string



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-02-01

import re
from pathlib import Path

from . import structs
from .file import (
    FILES,
    DevSettingFile,
    DjmMySettingFile,
    MySetting2File,
    MySettingFile,
    SettingsFile,
)

RE_MYSETTING = re.compile(".*SETTING[0-9]?.DAT$")


def get_mysetting_paths(root, deep=False):
    files = list()
    root = Path(root)
    iteator = root.rglob("*") if deep else root.iterdir()
    for path in iteator:
        if path.is_file() and RE_MYSETTING.match(path.name):
            files.append(path)
    return files


def read_mysetting_file(path) -> SettingsFile:
    obj = FILES[str(Path(path).name)]
    return obj.parse_file(path)



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-02-01

"""Rekordbox My-Setting file handlers."""

import re
from collections.abc import MutableMapping

from construct import Struct

from . import structs

# fmt: off
CRC16_XMODEM_TABLE = [
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
    0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
    0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
    0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
    0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
    0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
    0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
    0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
    0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
    0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
    0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
    0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
    0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
    0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
    0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
    0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
    0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
    0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
    0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
    0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
    0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
    0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
    0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
    0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
    0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
    0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
    0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0,
]
# fmt: on

RE_INVALID_KEY = re.compile("[_u][0-9]?", flags=re.IGNORECASE)


def compute_checksum(data, struct):
    """Computes the CRC16 XModem checksum for My-Setting files.

    The checksum is calculated over the contents of the `data` field,
    except for `DJMSETTING.DAT` files where the checksum is calculated over all
    preceding bytes including the length fields.

    Parameters
    ----------
    data : bytes
        The binary filey contents of the My-Setting file for which the checksum is
        computed.
    struct : Struct
        The ``Struct`` of the My-Setting file. This is used for deciding which bytes
        are used for computing the checksum.

    Returns
    -------
    crc : int
        The calculated CRC16-XModem checksum.

    References
    ----------
    https://reveng.sourceforge.io/crc-catalogue/all.htm#crc.cat.crc-16-xmodem
    """
    start = 0 if struct == structs.DjmMySetting else 104
    crc = 0
    for byte in data[start:-4]:
        crc = ((crc << 8) & 0xFF00) ^ CRC16_XMODEM_TABLE[((crc >> 8) & 0xFF) ^ byte]
    return crc


def _is_valid_key(k: str):
    return not RE_INVALID_KEY.match(k)


class SettingsFile(MutableMapping):
    """Base class for the Rekordbox My-Setting file handler.

    The base class implements the getters and setter defined by the keys and
    default values in the ``defaults`` class attribute.
    The keys of the ``defaults`` dictionary, which has to be defined by the inheriting
    class, defines the valid attributes of the My-Setting file. The values are used
    as default values if a new (and empty) My-Setting file is initialized.
    """

    struct: Struct
    defaults: dict
    version: str = ""  # only used by DEVSETTING

    def __init__(self):
        super().__init__()
        self.parsed = None
        self._items = dict()

    @classmethod
    def parse(cls, data: bytes):
        """Parses the in-memory data of a Rekordbox settings binary file.

        Parameters
        ----------
        data : bytes
            The in-memory binary contents of a Rekordbox settings file.

        Returns
        -------
        self : SettingsFile
            The new instance with the parsed file content.
        """
        self = cls()
        self._parse(data)
        return self

    @classmethod
    def parse_file(cls, path: str):
        """Reads and parses a Rekordbox settings binary file.

        Parameters
        ----------
        path : str
            The path of a Rekordbox settings file which is used to read
            the file contents before parsing the binary data.

        Returns
        -------
        self : AnlzFile
            The new instance with the parsed file content.

        See Also
        --------
        SettingsFile.parse: Parses the data of a Rekordbox settings file.
        """
        with open(path, "rb") as fh:
            data = fh.read()
        return cls.parse(data)

    def _parse(self, data):
        parsed = self.struct.parse(data)
        keys = filter(_is_valid_key, parsed.data.keys())
        items = dict()
        for key in keys:
            items[key] = str(parsed.data[key])

        self.parsed = parsed
        self._items.update(items)

    def __len__(self):
        return len(self.defaults.keys())

    def __iter__(self):
        return iter(self.defaults)

    def __getitem__(self, key):
        try:
            return self._items[key]
        except KeyError:
            return self.defaults[key]

    def __setitem__(self, key, value):
        if key not in self.defaults.keys():
            raise KeyError(f"Key {key} not a valid field of {self.__class__.__name__}")
        self._items[key] = value

    def __delitem__(self, key):
        del self._items[key]

    def get(self, key, default=None):
        """Returns the value of a setting of the My-Setting file.

        If the key is not found in the My-Setting data, but it is present in the
        ``defaults`` class dictionary, that default value is used. Otherwise, the
        parameter ``default`` is used as default value.

        Parameters
        ----------
        key : str
            The key of the setting.
        default : Any, optional
            The default value returned if the setting does not exist in the
            My-Setting file data or the ``defaults`` dictionary.

        Returns
        -------
        value : Any
            The value of the setting.
        """
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def set(self, key, value):
        """Sets the value of a setting of the My-Setting file.

        Parameters
        ----------
        key : str
            The key of the setting.
        value : Any
            The new value for updating the setting.
        """
        self.__setitem__(key, value)

    def build(self):
        """Constructs the binary data for saving the My-Setting file.

        Returns
        -------
        byte_data : bytes
            The binary file contents fot eh My-Setting file.
        """
        # Copy defaults and update with cuirrent data
        items = self.defaults.copy()
        items.update(self._items)

        # Create file data
        file_items = {"data": items, "checksum": 0}
        if self.version:
            file_items["version"] = self.version

        # Compute and update checksum. For this the data has to be serialized twice:
        # Once for generating the checksum and another time for writing the data
        # with the updated checksum
        data = self.struct.build(file_items)
        checksum = compute_checksum(data, self.struct)
        file_items["checksum"] = checksum

        # Write data with updated checksum
        return self.struct.build(file_items)

    def save(self, path):
        """Save the contents of the My-Setting file object.

        Parameters
        ----------
        path : str
            The file path used for saving.

        See Also
        --------
        build: Constructs the binary data of the file.
        """
        data = self.build()
        with open(path, "wb") as fh:
            fh.write(data)

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class MySettingFile(SettingsFile):
    """Rekordbox `MYSETTING.DAT` file handler.

    See Also
    --------
    SettingsFile : Base class implementing getters and setter defined by the keys and
        default values in the ``defaults`` class attribute.
    """

    struct = structs.MySetting
    defaults = {
        "auto_cue": structs.AutoCue.on,
        "auto_cue_level": structs.AutoCueLevel.memory,
        "disc_slot_illumination": structs.DiscSlotIllumination.bright,
        "eject_lock": structs.EjectLock.unlock,
        "hotcue_autoload": structs.HotCueAutoLoad.on,
        "hotcue_color": structs.HotCueColor.off,
        "jog_mode": structs.JogMode.vinyl,
        "jog_ring_brightness": structs.JogRingBrightness.bright,
        "jog_ring_indicator": structs.JogRingIndicator.on,
        "language": structs.Language.english,
        "lcd_brightness": structs.LCDBrightness.three,
        "master_tempo": structs.MasterTempo.off,
        "needle_lock": structs.NeedleLock.lock,
        "on_air_display": structs.OnAirDisplay.on,
        "phase_meter": structs.PhaseMeter.type1,
        "play_mode": structs.PlayMode.single,
        "quantize": structs.Quantize.on,
        "quantize_beat_value": structs.QuantizeBeatValue.one,
        "slip_flashing": structs.SlipFlashing.on,
        "sync": structs.Sync.off,
        "tempo_range": structs.TempoRange.ten,
        "time_mode": structs.TimeMode.remain,
    }


class MySetting2File(SettingsFile):
    """Rekordbox `MYSETTING2.DAT` file handler.

    See Also
    --------
    SettingsFile : Base class implementing getters and setter defined by the keys and
        default values in the ``defaults`` class attribute.
    """

    struct = structs.MySetting2
    defaults = {
        "vinyl_speed_adjust": structs.VinylSpeedAdjust.touch,
        "jog_display_mode": structs.JogDisplayMode.auto,
        "pad_button_brightness": structs.PadButtonBrightness.three,
        "jog_lcd_brightness": structs.JogLCDBrightness.three,
        "waveform_divisions": structs.WaveformDivisions.phrase,
        "waveform": structs.Waveform.waveform,
        "beat_jump_beat_value": structs.BeatJumpBeatValue.sixteen,
    }


class DjmMySettingFile(SettingsFile):
    """Rekordbox `DJMMYSETTING.DAT` file handler.

    See Also
    --------
    SettingsFile : Base class implementing getters and setter defined by the keys and
        default values in the ``defaults`` class attribute.
    """

    struct = structs.DjmMySetting
    defaults = {
        "channel_fader_curve": structs.ChannelFaderCurve.linear,
        "cross_fader_curve": structs.CrossfaderCurve.fast_cut,
        "headphones_pre_eq": structs.HeadphonesPreEQ.post_eq,
        "headphones_mono_split": structs.HeadphonesMonoSplit.stereo,
        "beat_fx_quantize": structs.BeatFXQuantize.on,
        "mic_low_cut": structs.MicLowCut.on,
        "talk_over_mode": structs.TalkOverMode.advanced,
        "talk_over_level": structs.TalkOverLevel.minus_18db,
        "midi_channel": structs.MidiChannel.one,
        "midi_button_type": structs.MidiButtonType.toggle,
        "display_brightness": structs.MixerDisplayBrightness.five,
        "indicator_brightness": structs.MixerIndicatorBrightness.three,
        "channel_fader_curve_long": structs.ChannelFaderCurveLong.exponential,
    }


class DevSettingFile(SettingsFile):
    """Rekordbox `DEVSETTING.DAT` file handler.

    Warnings
    --------
    The data of the `DEVSETTING.DAT` file is not supported. Only the header can be
    parsed and written. This class is implemented for completness only.

    See Also
    --------
    SettingsFile : Base class implementing getters and setter defined by the keys and
        default values in the ``defaults`` class attribute.
    """

    struct = structs.DevSetting
    defaults = dict(entries="")


FILES = {
    "DEVSETTING.DAT": DevSettingFile,
    "DJMMYSETTING.DAT": DjmMySettingFile,
    "MYSETTING.DAT": MySettingFile,
    "MYSETTING2.DAT": MySetting2File,
}



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-02-01

import re
from pathlib import Path
from typing import Union

from . import structs
from .file import AnlzFile

RE_ANLZ = re.compile("ANLZ[0-9]{4}.(DAT|EXT|2EX)")


def is_anlz_file(path: Union[str, Path]) -> bool:
    """Checks if the name of a file matches the ANLZ file name pattern.

    Parameters
    ----------
    path : str or Path
        The file path to check

    Returns
    -------
    is_anlz : bool
        True if the file is an ANLZ file.

    Examples
    --------
    >>> is_anlz_file("ANLZ0000.DAT")
    True

    >>> is_anlz_file("/path/to/ANLZ0000.DAT")
    False

    >>> is_anlz_file("ANLZ.DAT")
    False
    """
    path = Path(path)
    if not path.exists() or not path.is_file():
        return False
    return bool(RE_ANLZ.match(path.name))


def get_anlz_paths(root: Union[str, Path]) -> dict:
    """Returns the paths of all existing ANLZ files in a directory.

    Parameters
    ----------
    root : str or Path
        The path of the directory containing ANLZ files.

    Returns
    -------
    anlz_paths : dict[str, Path]
        The file paths stored as dictionary with the type of ANLZ file as keys
        ("DAT", "EXT", "EX2")

    Examples
    --------
    >>> p = get_anlz_paths("directory/")
    >>> p["DAT"]
    directory/ANLZ0000.DAT
    """
    paths = {"DAT": None, "EXT": None, "2EX": None}
    for path in Path(root).iterdir():
        if RE_ANLZ.match(path.name):
            paths[path.suffix[1:].upper()] = path
    return paths


def walk_anlz_dirs(root_dir: Union[str, Path]):
    """Finds all ANLZ directory paths recursively.

    Parameters
    ----------
    root_dir : str or Path
        The path of the root directory.

    Yields
    ------
    anlz_dir : str
        The path of a directory containing ANLZ files
    """
    for path in Path(root_dir).rglob("*"):
        if path.is_dir():
            if any(is_anlz_file(f) for f in path.iterdir()):
                yield path


def walk_anlz_paths(root_dir: Union[str, Path]):
    """Finds all ANLZ directory paths and the containing file paths recursively.

    Parameters
    ----------
    root_dir : str or Path
        The path of the root directory.

    Yields
    ------
    anlz_dir : str
        The path of a directory containing ANLZ files.
    anlz_files : Sequence of str
        The file paths of the ANLZ files in `anlz_dir`.
    """
    root_dir = Path(root_dir)
    assert root_dir.exists()
    for anlz_dir in walk_anlz_dirs(root_dir):
        files = {k: path for k, path in get_anlz_paths(anlz_dir).items() if path}
        yield anlz_dir, files


def read_anlz_files(root: Union[str, Path] = "") -> dict:
    """Open all ANLZ files in the given root directory.

    Parameters
    ----------
    root : str or Path, optional
        The root directory of the ANLZ files to open.

    Returns
    -------
    anlz_files : dict
        The opened ANLZ files stored in a dict with the corresponding file paths
        as keys.
    """
    paths = get_anlz_paths(root)
    return {path: AnlzFile.parse_file(path) for path in paths.values() if path}



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-09-10

import xml.etree.cElementTree as xml
from datetime import datetime
from pathlib import Path

from ..config import get_config
from ..utils import pretty_xml


class MasterPlaylistXml:
    """Rekordbox v6 masterPlaylists6.xml file handler.

    Rekordbox stores some playlist information in the masterPlaylists6.xml file.
    Each playlist is represented by a <PLAYLIST> element, containing the following
    attributes:
    - Id: The playlist ID in hexadecimal format.
    - ParentId: The parent playlist ID in hexadecimal format. The root playlist has
    - Attributes: The type of playlist. 0 = normal, 1 = folder, 4 = smart playlist.
    - Timestamp: The last time the playlist was updated.
    - Lib_Type: ? (0 for palylists/folders)
    - CheckType: ? (always 0)
    """

    KEYS = ["Id", "ParentId", "Attributes", "Timestamp", "Lib_Type", "CheckType"]

    def __init__(self, path=None, db_dir=None):
        if path is None:
            if db_dir is None:
                db_dir = get_config("rekordbox6", "db_dir")
            path = Path(db_dir) / "masterPlaylists6.xml"

        tree = xml.parse(str(path))
        self.path = path
        self.root = tree.getroot()
        self.product = self.root.find("PRODUCT")
        self.playlists = self.root.find("PLAYLISTS")
        self._changed = False

    @property
    def version(self):
        return self.root.attrib["Version"]

    @property
    def automatic_sync(self):
        return self.root.attrib["AutomaticSync"]

    @property
    def rekordbox_version(self):
        return self.product.attrib["Version"]

    @property
    def modified(self):
        return self._changed

    def get_playlists(self):
        """Returns a list of the attributes of all playlist elements."""
        items = list()
        for playlist in self.playlists:
            items.append(playlist.attrib)
        return items

    def get(self, playlist_id):
        """Returns element attribs with the PlaylistID used in the `master.db` database.

        Parameters
        ----------
        playlist_id : str or int
            The playlist ID used in the main `master.db` database. This id is converted
            to hexadecimal format before searching.

        Returns
        -------
        playlist : dict
        """
        hex_id = f"{int(playlist_id):X}"
        element = self.playlists.find(f'.//NODE[@Id="{hex_id}"]')
        if element is None:
            return None
        attribs = dict(element.attrib)
        attribs["Attribute"] = int(attribs["Attribute"])
        attribs["Timestamp"] = datetime.fromtimestamp(int(attribs["Timestamp"]) / 1000)
        attribs["Lib_Type"] = int(attribs["Lib_Type"])
        attribs["CheckType"] = int(attribs["CheckType"])
        return attribs

    def add(
        self,
        playlist_id: str,
        parent_id: str,
        attribute: int,
        updated_at: datetime,
        lib_type: int = 0,
        check_type: int = 0,
    ):
        """Adds a new element with the PlaylistID used in the `master.db` database.

        Parameters
        ----------
        playlist_id : str or int
            The playlist ID used in the main `master.db` database. This id is converted
            to hexadecimal format before searching.
        parent_id : str or int, optional
            The parent playlist ID used in the main `master.db` database. This id is
            converted to hexadecimal format.
        attribute : int, optional
            The type of playlist. 0 = normal, 1 = folder, 4 = smart playlist.
        updated_at : datetime, optional
            The last time the playlist was updated.
        lib_type : int, optional
            The libarray type. It seems to be always 0 for playlists.
        check_type : int, optional
            The check type. It seems to be always 0.

        Returns
        -------
        element : xml.Element
            The newly created element.
        """
        hex_id = f"{int(playlist_id):X}"
        parent_id = f"{int(parent_id):X}" if parent_id != "root" else "0"
        timestamp = int(updated_at.timestamp() * 1000)
        attrib = {
            "Id": hex_id,
            "ParentId": parent_id,
            "Attribute": str(attribute),
            "Timestamp": str(timestamp),
            "Lib_Type": str(lib_type),
            "CheckType": str(check_type),
        }
        if self.playlists.find(f'.//NODE[@Id="{hex_id}"]') is not None:
            raise ValueError(f"Playlist with ID {playlist_id} ({hex_id}) exists.")
        element = xml.SubElement(self.playlists, "NODE", attrib=attrib)
        self._changed = True
        return element

    def remove(self, playlist_id):
        """Removes the element with the PlaylistID used in the `master.db` database.

        Parameters
        ----------
        playlist_id : str or int
            The playlist ID used in the main `master.db` database. This id is converted
            to hexadecimal format before searching.
        """
        hex_id = f"{int(playlist_id):X}"
        element = self.playlists.find(f'.//NODE[@Id="{hex_id}"]')
        if element is None:
            raise ValueError(f"Playlist with ID {playlist_id} ({hex_id}) not found.")
        self.playlists.remove(element)
        self._changed = True

    def update(
        self,
        playlist_id: str,
        parent_id: str = None,
        attribute: int = None,
        updated_at: datetime = None,
        lib_type: int = None,
        check_type: int = None,
    ):
        """Updates the element with the PlaylistID used in the `master.db` database.

        Parameters
        ----------
        playlist_id : str or int
            The playlist ID used in the main `master.db` database. This id is converted
            to hexadecimal format before searching.
        parent_id : str or int, optional
            The parent playlist ID used in the main `master.db` database. This id is
            converted to hexadecimal format.
        attribute : int, optional
            The type of playlist. 0 = normal, 1 = folder, 4 = smart playlist.
        updated_at : datetime, optional
            The last time the playlist was updated.
        lib_type : int, optional
            The libarray type. It seems to be always 0 for playlists.
        check_type : int, optional
            The check type. It seems to be always 0.
        """
        hex_id = f"{int(playlist_id):X}"
        element = self.playlists.find(f'.//NODE[@Id="{hex_id}"]')
        if element is None:
            raise ValueError(f"Playlist with ID {playlist_id} ({hex_id}) not found.")

        attribs = dict()
        if parent_id is not None:
            attribs["ParentId"] = f"{int(parent_id):X}" if parent_id != "root" else "0"
        if attribute is not None:
            attribs["Attribute"] = str(attribute)
        if updated_at is not None:
            attribs["Timestamp"] = str(int(updated_at.timestamp() * 1000))
        if lib_type is not None:
            attribs["Lib_Type"] = str(lib_type)
        if check_type is not None:
            attribs["CheckType"] = str(check_type)

        element.attrib.update(attribs)
        self._changed = True

    def to_string(self, indent=None):
        return pretty_xml(self.root, indent, encoding="utf-8")

    def save(self, path=None, indent=None):
        if path is None:
            path = self.path
        path = str(path)

        string = self.to_string(indent)
        with open(path, "w") as fh:
            fh.write(string)
        self._changed = False
